package ru.ifmo.vizi.ctree;

import ru.ifmo.vizi.ctree.ui.*;
import ru.ifmo.vizi.ctree.widgets.*;
import ru.ifmo.vizi.ctree.timer.*;

import ru.ifmo.vizi.base.ui.*;
import ru.ifmo.vizi.base.*;
import ru.ifmo.vizi.base.widgets.*;

import java.awt.*;
import java.awt.event.*;
import java.util.Stack;

public class CTreeVisualizer extends Base implements MouseListener, MouseMotionListener, KeyListener
{
    public final static int STANDARD = 0;
    public final static int PREZERO = 11;
    public final static int ZERO = 1;
    public final static int TRY_TREE = 2;
    public final static int BUILD_SCC = 3;
    public final static int NEW_GRAPH = 4;
    public final static int PRE_BUILD = 5;
    public final static int OPEN_COMPONENTS = 6;
    public final static int PRE_TREE = 7;
    public final static int BUILD_TREE = 8;
    public final static int FINISH = 9;
    public final static int ERROR = 10;

    private final Example DEFAULT_EXAMPLE = new Example(3, new double[][] {{Double.NaN, 2, 2},
                                                                           {2, Double.NaN, 1},
                                                                           {2, 1, Double.NaN}},
                                                               "Default Example");

    private final CTree auto;
    private final CTree.Data data;
    private final Frame forefather;

    private String minsString = "mins";

    // relative sizes and places of drawing items.
    private double gRadius = 0.75;
    private double vRadius = 0.25;
    private double tSize = 0.9;
    private double split = 0.5;
    private double gsplit = 0.5;
    private boolean gAspectStatus = true;
    private double gAspect = 1.0;
    private boolean tAspectStatus = true;
    private double tAspect = 1.0;
    private double minsplit = 0.0;
    private double maxsplit = 1.0;

    private ShapeStyle[] tableStyle;
    private ShapeStyle[] vertexStyle;
    private ShapeStyle[] arrowStyle;

    private Stack exampleSet;

    public int maxVertices = 8;
    public int minVertices = 3;

    private int nVertices;
    private StringPanel examples;
    private Rect edit;
    private StringBuffer buf;
    private int maxChars = 3;
    private Panel autoPane;
    private Panel controlsPane;
    private SpinPanel vertPane;

    private boolean editor;
    private MultiButton editButton;
    private HintedCheckbox showcheck;
    private HintedCheckbox modecheck;

    private int viziMode;

    // variables for splitter.
    private int offset;
    private boolean moveSplit;
    private Graph moveGraph;
    private Rect gsplitRect;
    private int agsplit;

    // components for adding new edge.
    private Ellipse arrowBegin;
    private Ellipse arrowEnd;
    private Arrow newArrow;

    // edit frame.
    //private EditFrame frame;
    private EditDialog frame;

    // graphs.
    private Graph mainGraph;
    private Graph tempGraph;
    private final Stack tempGraphs = new Stack();
    private CTreeTimer timer;
    private DelayPanel delayPanel;
    private boolean timerActive;
    private boolean collapseSCC;

    public CTreeVisualizer(VisualizerParameters parameters)
    {
        super(parameters);
        auto = new CTree(locale);
        data = auto.d;
        data.visualizer = this;
        data.funs = new CTreeFuns(auto);
        forefather = parameters.getForefather();

        createInterface(auto);

        clientPane.addMouseListener(this);
        clientPane.addKeyListener(this);
        clientPane.addMouseMotionListener(this);

        tableStyle = ShapeStyle.loadStyleSet(config, "table");
        vertexStyle = ShapeStyle.loadStyleSet(config, "vertex");
        arrowStyle = ShapeStyle.loadStyleSet(config, "arrow");

        exampleSet = new Stack();
        for (int i = 0; i < config.getInteger("examples", 0); i++)
        {
            Example ex = loadExample("example" + i, DEFAULT_EXAMPLE);
            exampleSet.push(ex);
            examples.addString(ex.getName());
        }

        maxVertices = config.getInteger("vertices-value-max", maxVertices);
        minVertices = config.getInteger("vertices-value-min", minVertices);

        minsString = config.getParameter("minsString", minsString);
        gRadius = config.getDouble("gRadius", gRadius);
        vRadius = config.getDouble("vRadius", vRadius);
        tSize = config.getDouble("tSize", tSize);
        split = config.getDouble("split", split);
        gsplit = config.getDouble("gsplit", gsplit);
        gAspectStatus = config.getBoolean("gAspectStatus", gAspectStatus);
        gAspect = config.getDouble("gAspect", gAspect);
        tAspectStatus = config.getBoolean("tAspectStatus", tAspectStatus);
        tAspect = config.getDouble("tAspect", tAspect);
        minsplit = config.getDouble("min-split", minsplit);
        maxsplit = config.getDouble("max-split", maxsplit);

        gsplitRect = new Rect(new ShapeStyle[]{new ShapeStyle(config, "gsplitRectStyle", tableStyle[0])});
        gsplitRect.setVisible(false);
        gsplitRect.setCursor(new Cursor(Cursor.W_RESIZE_CURSOR));

        mainGraph = new Graph(auto, config, minsString, gRadius, gAspectStatus,
                              gAspect, tAspectStatus, tAspect,
                              vRadius, tSize, split, minsplit, maxsplit,
                              tableStyle, vertexStyle, arrowStyle);

        moveSplit = false;
        boolean show = config.getBoolean("showcheck", true);
        showcheck.setState(show);

        boolean mode = config.getBoolean("modecheck", true);
        modecheck.setState(mode);
        if (mode) viziMode = 0;
        else viziMode = 1;

        timerActive = false;

        setExample((Example) exampleSet.elementAt(0));
        edit = null;
        maxChars = config.getInteger("maxChars", maxChars);

        //frame = new EditFrame(config, this);
        frame = new EditDialog(config, this, forefather);
        editor = false;
    }

    public Component createControlsPane()
    {
        controlsPane = new Panel(new BorderLayout());

        autoPane = new Panel();
        AutoControlsPane acp = new AutoControlsPane(config, auto, forefather, false);

        delayPanel = findDelayPanel(acp);
        autoPane.add(acp);
        controlsPane.add(autoPane, BorderLayout.NORTH);

        Panel specPanel = new Panel();

        specPanel.add(examples = new StringPanel(config, "examples", locale)
        {
            public void click(double value)
            {
                setExample((Example) exampleSet.elementAt(getCurrentIndex()));
            }
        });

        specPanel.add(editButton = new MultiButton(config, new String[]{"edit-button", "start-button"})
        {
            public int click(int state)
            {
                switch (state)
                {
                    case 0:
                        setEditor(true);
                        return 1;
                    case 1:
                        setEditor(false);
                        return 0;
                    default:
                        return 0;
                }
            }
        });

        if (config.getBoolean("has-save", true))
        {
            specPanel.add(new HintedButton(config, "save-button")
            {
                public void click()
                {
                    int step = 0;
                    while (!auto.isAtStart())
                    {
                        auto.stepBackward(0);
                        step++;
                    }
                    frame.loadExample(new Example(data.n, data.graph, examples.getCurrentString()), step);
                    for (int i = 0; i < step; i++)
                    {
                        auto.stepForward(0);
                    }
                    frame.center();
                    frame.show();
                }
            });
        }
        specPanel.add(showcheck = new HintedCheckbox(config, "show-check", locale)
        {
            public void click()
            {
                if (showcheck.getState())
                {
                    mainGraph.setTableVisible(true);
                    if (tempGraph != null) tempGraph.setTableVisible(true);
                }
                else
                {
                    mainGraph.setTableVisible(false);
                    if (tempGraph != null) tempGraph.setTableVisible(false);
                }
            }
        });
        specPanel.add(modecheck = new HintedCheckbox(config, "vizi-mode", locale)
        {
            public void click()
            {
                viziMode = (viziMode + 1) % 2;
            }
        });

        controlsPane.add(specPanel, BorderLayout.SOUTH);
        return controlsPane;
    }

    public DelayPanel findDelayPanel(AutoControlsPane p)
    {
        Component[] c = p.getComponents();
        for (int i = 0; i < c.length; i++)
        {
            if (c[i] instanceof DelayPanel) return (DelayPanel) c[i];
        }
        return null;
    }

    public void clearTimer()
    {
        timer.stop();
        timer = null;
    }

    public void layoutClientPane(int width, int height)
    {
        reshapePane();
    }

    public void copyGraph(boolean isnew)
    {
        if (isnew)
        {
            tempGraph = new Graph(mainGraph);
            tempGraphs.push(tempGraph);
        }
        else
        {
            tempGraph = (Graph) tempGraphs.peek();
        }
        if (viziMode == 0) return;      // show two graphs
        timerActive = isnew;
    }

    public void clearGraph()
    {
        tempGraph = null;
    }

    public void popGraph()
    {
        tempGraphs.pop();
        tempGraph = null;
    }

    public void setCollapse(boolean c)
    {
        collapseSCC = c;
    }

    public void reshapePane()
    {
        if (timer != null)
        {
            timer.stop();
            timer = null;
        }
        int width = clientPane.getSize().width;
        int height = clientPane.getSize().height;
        agsplit = (int) (width * gsplit);
        clientPane.removeAll();
        mainGraph.updateGraph(auto);
        if (tempGraph == null)
        {
            mainGraph.reshapeGraph(0, 0, width, height);
            mainGraph.addTo(clientPane);
            gsplitRect.setVisible(false);
        }
        else if (viziMode == 0)
        {
            mainGraph.reshapeGraph(0, 0, agsplit, height);
            tempGraph.reshapeGraph(0, agsplit, width - agsplit, height);
            mainGraph.addTo(clientPane);
            tempGraph.addTo(clientPane);
            gsplitRect.setVisible(true);
        }
        else if (timerActive)
        {
            mainGraph.reshapeGraph(0, 0, width, height);
            tempGraph.reshapeGraph(0, 0, width, height);
            int delay = delayPanel.getIntValue() / (CTreeTimer.STEPS + 1);
            Graph viewGraph;
            if (collapseSCC) viewGraph = tempGraph;
            else viewGraph = mainGraph;
            timer = new CTreeTimer(this, tempGraph, mainGraph, viewGraph, data, delay);
            timer.start();
            viewGraph.addTo(clientPane);
            gsplitRect.setVisible(false);
        }
        else
        {
            mainGraph.reshapeGraph(0, 0, width, height);
            mainGraph.addTo(clientPane);
            gsplitRect.setVisible(false);
        }

        gsplitRect.setBounds(agsplit - 2, 0, 4, height);
        clientPane.add(gsplitRect);
        if (newArrow != null)
        {
            clientPane.add(newArrow);
        }
    }

    public void updatePane(int flag)
    {
        switch (flag)
        {
            case NEW_GRAPH:
            case OPEN_COMPONENTS:
                //if (timer != null) timer.start();
                break;
            default:
                if (timer != null)
                {
                    timer.stop();
                    timer = null;
                    mainGraph.reshapeGraph();
                }
        }
        mainGraph.updateGraph(auto);
        mainGraph.updateTable(flag);
        if (tempGraph != null)
        {
            tempGraph.updateTable(flag - 1);
        }
    }

    public void update()
    {
        clientPane.invalidate();
        update(true);
    }

    public Example loadExample(String prefix, Example def)
    {
        String s = config.getParameter(prefix);
        String suffix = "-" + locale.getLanguage();
        Example e = new Example();
        try
        {
            e.readExample(s, config, this, suffix);
        }
        catch (Exception ex)
        {
            System.out.println(ex.getMessage());
        }
        return e;
    }

    public void addExample(Example e)
    {
        exampleSet.push(e);
        examples.addString(e.getName());
        examples.setCurrentIndex(exampleSet.size() - 1);
    }

    public void deleteExample(int index)
    {
        if (exampleSet.size() == 1) return;
        exampleSet.removeElementAt(index);
        examples.removeString(index);
        if (index < exampleSet.size()) setExample((Example) exampleSet.elementAt(index));
        else setExample((Example) exampleSet.elementAt(index - 1));
    }

    public void setExample(Example e)
    {
        while (!auto.isAtStart()) auto.stepBackward(0);
        data.n = e.getNumber();
        data.graph = e.getMatrix();
        setVertices(data.n);
        int step = e.getStep();
        if (step > 0)
        {
            setEditor(false);
        }
        while (step != 0)
        {
            auto.stepForward(0);
            step--;
        }
    }

    public void setVertices(int n)
    {
        nVertices = n;
        if (vertPane != null) vertPane.setValue(n);
        updateAuto(n);
        mainGraph.updateGraph(auto);
        reshapePane();
        mainGraph.updateTable(STANDARD);
        update();
    }

    public void updateAuto(int n)
    {
        double[][] matrix = new double[n][n];
        for (int i = 0; i < n; i++)
        {
            for (int j = 0; j < n; j++)
            {
                if ((i < data.n) && (j < data.n))
                {
                    matrix[i][j] = data.graph[i][j];
                }
                else
                {
                    matrix[i][j] = Double.NaN;
                }
            }
        }
        data.n = n;
        data.graph = matrix;
    }

    public void setEditor(boolean ed)
    {
        editor = ed;
        autoPane.removeAll();
        if (ed)
        {
            while (!auto.isAtStart()) auto.stepBackward(0);
            autoPane.add(vertPane = new SpinPanel(config, "vertices")
            {
                public void click(double value)
                {
                    setVertices((int) value);
                }
            });
            autoPane.add(new HintedButton(config, "clear-button")
            {
                public void click()
                {
                    clearEdges();
                }
            });
            setVertices(nVertices);
            editButton.setState(1);
        }
        else
        {
            vertPane = null;
            AutoControlsPane acp = new AutoControlsPane(config, auto, forefather, false);
            delayPanel = findDelayPanel(acp);
            autoPane.add(acp);
            setEdgeLength();
            edit = null;
            newArrow = null;
            arrowBegin = null;
            arrowEnd = null;
            editButton.setState(0);
        }
        controlsPane.doLayout();
        controlsPane.validate();
        mainGraph.updateTable(STANDARD);
        update();
    }

    public void mouseExited(MouseEvent e)
    {
    }

    public void mouseEntered(MouseEvent e)
    {
    }

    public void mouseClicked(MouseEvent e)
    {
        if (editor == false) return;
        Point p = e.getPoint();
        if (edit != null)
        {
            setEdgeLength();
            mainGraph.updateTable(STANDARD);
            edit = null;
        }
        Component c = clientPane.getComponentAt(p);
        if (c instanceof Rect) edit = (Rect) c;
        c = clientPane.getComponentAt(p);
        if (c instanceof Rect) edit = (Rect) c;
        if (edit == null) return;
        int index = mainGraph.getTable().indexOf(edit);
        int i = index / data.n;
        int j = index % data.n;
        if ((i == j) || (index >= data.n * data.n))
        {
            edit = null;
            return;
        }
        edit.setStyle(4);
        index = i * (data.n - 1) + j;
        if (j > i) index--;
        ((Arrow) mainGraph.getArrows().elementAt(index)).setStyle(4);
        buf = new StringBuffer();
        update();
    }

    public void mousePressed(MouseEvent e)
    {
        if (editor == false) return;
        if (newArrow == null)
        {
            Point p = e.getPoint();
            Component c = clientPane.getComponentAt(p);
            if (c instanceof Ellipse)
            {
                arrowBegin = (Ellipse) c;
            }
            if (arrowBegin != null)
            {
                 while (!auto.isAtStart())
                {
                    auto.stepBackward(0);
                }
                newArrow = new Arrow(arrowStyle);
                newArrow.setArrow(p.x, p.y, p.x, p.y);
                clientPane.add(newArrow);
            }
        }
        update();
    }

    public void mouseReleased(MouseEvent e)
    {
        if (newArrow == null) return;
        Point p = e.getPoint();
        Component c = clientPane.getComponentAt(p);
        if (c instanceof Ellipse)
        {
            arrowEnd = (Ellipse) c;
        }
        if ((arrowEnd != null) && (arrowBegin != null))
        {
            int i = mainGraph.getVertices().indexOf(arrowBegin);
            int j = mainGraph.getVertices().indexOf(arrowEnd);
            if (i != j)
            {
                if (Double.isNaN(data.graph[i][j])) data.graph[i][j] = 0;
                if (edit != null)
                {
                    setEdgeLength();
                    mainGraph.updateTable(STANDARD);
                }
                edit = (Rect) mainGraph.getTable().elementAt(i * data.n + j);
                edit.setStyle(4);
                int index = i * (data.n - 1) + j;
                if (j > i) index--;
                Arrow arr = (Arrow) mainGraph.getArrows().elementAt(index);
                arr.setStyle(4);
                arr.setVisible(true);
                buf = new StringBuffer();
            }
        }
        clientPane.remove(newArrow);
        newArrow = null;
        arrowBegin = null;
        arrowEnd = null;
        update();
    }

    public void mouseDragged(MouseEvent e)
    {
        if (newArrow != null)
        {
            Point p = e.getPoint();
            newArrow.setEnd(p.x, p.y);
            return;
        }
        if (moveSplit)
        {
            if (moveGraph != null)
            {
                moveGraph.setSplit(e.getX() + offset);
            }
            else
            {
                agsplit = e.getX() + offset;
                gsplit = (double) agsplit / clientPane.getSize().width;
                reshapePane();
                update();
            }
        }
    }

    public void mouseMoved(MouseEvent e)
    {
        if (!showcheck.getState()) return;
        if ((e.getX() >= mainGraph.getSplit() - 2) && (e.getX() <= mainGraph.getSplit() + 2))
        {
            offset = mainGraph.getSplit() - e.getX();
            moveSplit = true;
            moveGraph = mainGraph;
            return;
        }
        if (tempGraph != null)
        {
            if ((e.getX() >= tempGraph.getSplit() - 2) && (e.getX() <= tempGraph.getSplit() + 2))
            {
                offset = tempGraph.getSplit() - e.getX();
                moveSplit = true;
                moveGraph = tempGraph;
                return;
            }
        }
        if ((e.getX() >= agsplit - 2) && (e.getX() <= agsplit + 2))
        {
            offset = agsplit - e.getX();
            moveSplit = true;
            moveGraph = null;
            return;
        }
        moveSplit = false;
    }

    public void setEdgeLength()
    {
        if (edit == null) return;
        int i,j;
        int index = mainGraph.getTable().indexOf(edit);
        i = index / data.n;
        j = index % data.n;
        String s = new String(buf);
        if (s.length() != 0)
        {
            try
            {
                data.graph[i][j] = Double.valueOf(s).doubleValue();
                if (data.graph[i][j] < 0) throw new NumberFormatException();
            }
            catch (NumberFormatException ex)
            {
                data.graph[i][j] = Double.NaN;
            }
        }
        mainGraph.updateGraph(auto);
    }

    public void clearEdges()
    {
        for (int i = 0; i < data.n; i++)
        {
            for (int j = 0; j < data.n; j++)
            {
                data.graph[i][j] = Double.NaN;
            }
        }
        mainGraph.updateGraph(auto);
        mainGraph.updateTable(STANDARD);
        update();
    }

    public void setEdgeToNaN()
    {
        int i,j;
        int index = mainGraph.getTable().indexOf(edit);
        i = index / data.n;
        j = index % data.n;
        data.graph[i][j] = Double.NaN;
        mainGraph.updateGraph(auto);
    }

    public void keyTyped(KeyEvent e)
    {
        if (edit == null) return;
        switch (e.getKeyChar())
        {
            case '\n':
                setEdgeLength();
                mainGraph.updateTable(STANDARD);
                edit = null;
                update();
                return;
            case '\b':
                if (buf.length() > 0) buf.setLength(buf.length() - 1);
                break;
            case 27:
                mainGraph.updateTable(STANDARD);
                edit = null;
                break;
            default:
                if (buf.length() < maxChars)
                {
                    buf.append(e.getKeyChar());
                }
        }
        edit.setMessage(new String(buf));
        edit.adjustFontSize();
        update();
    }

    public void keyPressed(KeyEvent e)
    {
        switch (e.getKeyChar())
        {
            case 127:
                if (edit == null) break;
                setEdgeToNaN();
                mainGraph.updateTable(STANDARD);
                edit = null;
                update();
                break;
            default:
        }
    }

    public void keyReleased(KeyEvent e)
    {
    }


}